import os
from typing import Dict, List
import shutil
import subprocess
import json
import vdf
import requests
import threading
from steam.utils.appcache import parse_appinfo

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QMessageBox, QApplication

from pupgui2.constants import LOCAL_AWACY_GAME_LIST, PROTONDB_API_URL
from pupgui2.constants import STEAM_STL_INSTALL_PATH, STEAM_STL_CONFIG_PATH, STEAM_STL_SHELL_FILES, STEAM_STL_FISH_VARIABLES
from pupgui2.datastructures import SteamApp, AWACYStatus, BasicCompatTool, CTType


_cached_app_list = []
_cached_steam_ctool_id_map = None


def get_steam_app_list(steam_config_folder: str, cached=False) -> List[SteamApp]:
    """
    Returns a list of installed Steam apps and optionally game names and the compatibility tool they are using
    steam_config_folder = e.g. '~/.steam/root/config'
    Return Type: List[SteamApp]
    """
    global _cached_app_list

    if cached and _cached_app_list != []:
        return _cached_app_list

    libraryfolders_vdf_file = os.path.join(os.path.expanduser(steam_config_folder), 'libraryfolders.vdf')
    config_vdf_file = os.path.join(os.path.expanduser(steam_config_folder), 'config.vdf')

    apps = []
    app_ids_str = []

    try:
        v = vdf.load(open(libraryfolders_vdf_file))
        c = vdf.load(open(config_vdf_file)).get('InstallConfigStore').get('Software').get('Valve').get('Steam').get('CompatToolMapping')
        for fid in v.get('libraryfolders'):
            if 'apps' not in v.get('libraryfolders').get(fid):
                continue
            fid_path = v.get('libraryfolders').get(fid).get('path')
            if fid == '0':
                fid_path = os.path.join(fid_path, 'steamapps', 'common')
            for appid in v.get('libraryfolders').get(fid).get('apps'):
                app = SteamApp()
                app.app_id = int(appid)
                app.libraryfolder_id = fid
                app.libraryfolder_path = fid_path
                if ct := c.get(appid):
                    app.compat_tool = ct.get('name')
                apps.append(app)
                app_ids_str.append(str(appid))
            apps = update_steamapp_info(steam_config_folder, apps)
            apps = update_steamapp_awacystatus(apps)
    except Exception as e:
        print('Error: Could not get a list of all Steam apps:', e)

    _cached_app_list = apps
    return apps


def get_steam_game_list(steam_config_folder: str, compat_tool='', cached=False) -> List[SteamApp]:
    """
    Returns a list of installed Steam games and which compatibility tools they are using.
    Specify compat_tool to only return games using the specified tool.
    Return Type: List[SteamApp]
    """
    apps = get_steam_app_list(steam_config_folder, cached=cached)

    return [app for app in apps if app.app_type == 'game' and (compat_tool == '' or app.compat_tool == compat_tool)]


def get_steam_ctool_list(steam_config_folder: str, only_proton=False, cached=False) -> List[SteamApp]:
    """
    Returns a list of installed Steam compatibility tools (official tools).
    Return Type: List[SteamApp]
    """
    ctools = []
    apps = get_steam_app_list(steam_config_folder, cached=cached)
    ctool_map = _get_steam_ctool_info(steam_config_folder)

    for app in apps:
        if ct := ctool_map.get(app.app_id):
            app.ctool_name = ct.get('name')
            app.ctool_from_oslist = ct.get('from_oslist')
            if only_proton and ct.get('from_oslist') != 'windows':
                continue
            ctools.append(app)

    return ctools


def get_steam_acruntime_list(steam_config_folder: str, cached=False) -> List[BasicCompatTool]:
    """
    Returns a list of installed Steam Proton anticheat(EAC/BattlEye) Runtimes.
    Return Type: List[BasicCompatTool]
    """
    runtimes = []
    apps = get_steam_app_list(steam_config_folder, cached=cached)

    for app in apps:
        if app.app_type == 'acruntime':
            ct = BasicCompatTool(app.game_name, app.libraryfolder_path, '', CTType.STEAM_RT)
            runtimes.append(ct)

    return runtimes


def _get_steam_ctool_info(steam_config_folder: str) -> Dict[str, Dict[str, str]]:
    """
    Returns a dict that maps the compatibility tool appid to tool info (name e.g. 'proton_7' and from_oslist)
    Return Type: Dict[str, dict]
        Contents: appid str -> {'name', 'from_oslist'}
    """
    global _cached_steam_ctool_id_map

    if _cached_steam_ctool_id_map is not None:
        return _cached_steam_ctool_id_map

    appinfo_file = os.path.join(os.path.expanduser(steam_config_folder), '../appcache/appinfo.vdf')
    appinfo_file = os.path.realpath(appinfo_file)

    ctool_map = {}
    compat_tools = None
    try:
        with open(appinfo_file, 'rb') as f:
            header, apps = parse_appinfo(f)
            for steam_app in apps:
                if steam_app.get('appid') == 891390:
                    compat_tools = steam_app.get('data').get('appinfo').get('extended').get('compat_tools')
                    break
    except:
        pass
    finally:
        for t in compat_tools:
            ctool_map[compat_tools.get(t).get('appid')] = {'name': t, 'from_oslist': compat_tools.get(t).get('from_oslist')}

    _cached_steam_ctool_id_map = ctool_map
    return ctool_map


def update_steamapp_info(steam_config_folder: str, steamapp_list: List[SteamApp]) -> List[SteamApp]:
    """
    Get Steam game names and information for provided SteamApps
    Return Type: List[SteamApp]
    """
    appinfo_file = os.path.join(os.path.expanduser(steam_config_folder), '../appcache/appinfo.vdf')
    appinfo_file = os.path.realpath(appinfo_file)
    sapps = {app.get_app_id_str(): app for app in steamapp_list}
    cnt = 0
    try:
        ctool_map = _get_steam_ctool_info(steam_config_folder)
        with open(appinfo_file, 'rb') as f:
            header, apps = parse_appinfo(f)
            for steam_app in apps:
                appid_str = str(steam_app.get('appid'))
                if a := sapps.get(appid_str):
                    try:
                        a.game_name = steam_app.get('data').get('appinfo').get('common').get('name')
                    except:
                        a.game_name = ''
                    try:
                        a.deck_compatibility = steam_app.get('data').get('appinfo').get('common').get('steam_deck_compatibility')
                    except:
                        pass
                    if steam_app.get('appid') not in ctool_map and 'steamworks' not in a.game_name.lower() \
                        and not ('Proton' in a.game_name and 'Runtime' in a.game_name):
                        a.app_type = 'game'
                    elif a.game_name.startswith('Proton') and a.game_name.endswith('Runtime'):
                        a.app_type = 'acruntime'
                    elif 'Steam Linux Runtime' in a.game_name:
                        a.app_type = 'runtime'
                    cnt += 1
                if cnt == len(sapps):
                    break
    except:
        pass
    return list(sapps.values())


def update_steamapp_awacystatus(steamapp_list: List[SteamApp]) -> List[SteamApp]:  # Download file in thread on start...
    """
    Set the areweanticheatyet.com for the games.
    Return Type: List[SteamApp]
    """
    if not os.path.exists(LOCAL_AWACY_GAME_LIST):
        return steamapp_list

    try:
        with open(LOCAL_AWACY_GAME_LIST, 'r') as f:
            gm = {g.get('name'): g.get('status') for g in json.load(f)}

        for app in steamapp_list:
            if app.game_name != '' and app.game_name in gm:
                status = gm[app.game_name]
                if status == 'Supported':
                    app.awacy_status = AWACYStatus.ASUPPORTED
                elif status == 'Planned':
                    app.awacy_status = AWACYStatus.PLANNED
                elif status == 'Running':
                    app.awacy_status = AWACYStatus.RUNNING
                elif status == 'Broken':
                    app.awacy_status = AWACYStatus.BROKEN
                elif status == 'Denied':
                    app.awacy_status = AWACYStatus.DENIED
    except Exception as e:
        print('Error updating the areweanticheatyet.com status:', e)
        return steamapp_list

    return steamapp_list


def get_protondb_status_thread(game: SteamApp, signal: Signal) -> None:
    """ Downloads the ProtonDB.com status and calls the Qt Signal "signal" when done. Use with "get_protondb_status"!"""
    try:
        json_url = PROTONDB_API_URL.format(game_id=str(game.app_id))
        r = requests.get(json_url)
        if r.status_code == 200:
            game.protondb_summary = r.json()
        signal.emit(game)
    except Exception as e:
        print('Error getting the protondb.com status:', e)
        signal.emit(None)


def get_protondb_status(game: SteamApp, signal: Signal) -> None:
    """ Downloads the ProtonDB.com status in a separate threads. When done the Qt Signal "signal" is called """
    t = threading.Thread(target=get_protondb_status_thread, args=(game, signal))
    t.start()


def steam_update_ctool(game: SteamApp, new_ctool=None, steam_config_folder='') -> bool:
    """
    Change compatibility tool for 'game_id' to 'new_ctool' in Steam config vdf
    Return Type: bool
    """
    config_vdf_file = os.path.join(os.path.expanduser(steam_config_folder), 'config.vdf')
    if not os.path.exists(config_vdf_file):
        return False

    game_id = game.app_id

    try:
        d = vdf.load(open(config_vdf_file))
        c = d.get('InstallConfigStore').get('Software').get('Valve').get('Steam').get('CompatToolMapping')

        if str(game_id) in c:
            if new_ctool is None:
                c.pop(str(game_id))
            else:
                c.get(str(game_id))['name'] = str(new_ctool)
        else:
            c[str(game_id)] = {"name": str(new_ctool), "config": "", "priority": "250"}

        vdf.dump(d, open(config_vdf_file, 'w'), pretty=True)
    except Exception as e:
        print('Error, could not update Steam compatibility tool to', new_ctool, 'for game',game_id, ':',
              e, ', vdf:', config_vdf_file)
        return False
    return True


def steam_update_ctools(games: Dict[SteamApp, str], steam_config_folder='') -> bool:
    """
    Change compatibility tool for multiple games in Steam config vdf.
    Return Type: bool
    """
    config_vdf_file = os.path.join(os.path.expanduser(steam_config_folder), 'config.vdf')
    if not os.path.exists(config_vdf_file):
        return False

    try:
        d = vdf.load(open(config_vdf_file))
        c = d.get('InstallConfigStore').get('Software').get('Valve').get('Steam').get('CompatToolMapping')

        for game, new_ctool in games.items():
            game_id = game.app_id
            if str(game_id) in c:
                if new_ctool is None:
                    c.pop(str(game_id))
                else:
                    c.get(str(game_id))['name'] = str(new_ctool)
            else:
                c[str(game_id)] = {"name": str(new_ctool), "config": "", "priority": "250"}

        vdf.dump(d, open(config_vdf_file, 'w'), pretty=True)
    except Exception as e:
        print('Error, could not update Steam compatibility tools:', e, ', vdf:', config_vdf_file)
        return False
    return True


def is_steam_running() -> bool:
    """
    Returns True if the Steam client is running, False otherwise
    Return Type: bool
    """
    try:
        procs = os.listdir('/proc')
        for proc in procs:
            exe = os.path.join('/proc', proc, 'exe')
            if os.path.exists(exe) and 'steam' in os.readlink(exe):
                return True
    except:
        pass
    return False


get_fish_user_paths = lambda mfile: ([line.strip() for line in mfile.readlines() if 'fish_user_paths' in line] or ['SETUVAR fish_user_paths:\\x1d'])[0].split('fish_user_paths:')[1:][0].split('\\x1e')


def get_external_steamtinkerlaunch_intall(compat_folder):

    symlink_path = os.path.join(compat_folder, 'steamtinkerlaunch')
    return os.path.dirname(os.readlink(symlink_path)) if os.path.exists(symlink_path) and os.readlink(symlink_path) != os.path.join(STEAM_STL_INSTALL_PATH, 'prefix', 'steamtinkerlaunch') else None


def remove_steamtinkerlaunch(compat_folder='', remove_config=True, ctmod_object=None) -> bool:
    """
    Removes SteamTinkerLaunch from system by removing the downloaad, removing from path
    removing config files at `$HOME/.config/steamtinkerlaunch`.
    
    Returns True if successfully removed.
    Reutrn Type: bool
    """

    try:
        os.chdir(os.path.expanduser('~'))

        # If the Steam Deck/ProtonUp-Qt installation path doesn't exist
        # Adding `prefix` to path to be especially sure the user didn't just make an `stl` folder
        #
        # STL script is always named `steamtinkerlaunch`    
        stl_symlink_path = get_external_steamtinkerlaunch_intall(compat_folder)

        if os.path.exists(compat_folder):
            print('Removing SteamTinkerLaunch compatibility tool...')
            shutil.rmtree(compat_folder)
            if shutil.which('steamtinkerlaunch'):
                subprocess.run(['steamtinkerlaunch', 'compat', 'del'])

        print('Removing SteamTinkerLaunch installation...')
        if stl_symlink_path:
            # If STL symlink isn't a regular install, try to remove if we can write to its install folder
            if os.access(stl_symlink_path, os.W_OK):
                shutil.rmtree(stl_symlink_path)
                print('Removed SteamTinkerLaunch installation folder pointed to by symlink')
            else:
                # If we can't remove the actual installation folder, tell the user to remove it themselves and continue with the rest of the uninstallation
                mb_title = QApplication.instance().translate('steamutil.py', 'Unable to Remove SteamTinkerLaunch')
                mb_text = QApplication.instance().translate(
                    'steamutil.py',
                    'Access to SteamTinkerLaunch installation folder at \'{STL_SYMLINK_PATH}\' was denied, please remove this folder manually.\n\nThe uninstallation will continue.'
                ).format(STL_SYMLINK_PATH=stl_symlink_path)
                if ctmod_object and hasattr(ctmod_object, 'message_box_message'):
                    ctmod_object.message_box_message.emit(mb_title, mb_text, QMessageBox.Icon.Warning)
                else:
                    mb = QMessageBox()
                    mb.setWindowTitle(mb_title)
                    mb.setText(mb_text)
                    mb.exec()

                print(f'Error: SteamTinkerLaunch is installed to {stl_symlink_path}, ProtonUp-Qt cannot modify this folder. Folder must be removed manually.')
        elif os.path.exists(STEAM_STL_INSTALL_PATH):
            # Regular Steam Deck/ProtonUp-Qt installation structure
            if os.path.exists('/.flatpak-info'):
                if os.path.exists(os.path.join(STEAM_STL_INSTALL_PATH, 'prefix')):
                    shutil.rmtree(os.path.join(STEAM_STL_INSTALL_PATH, 'prefix'))
            else:
                shutil.rmtree(STEAM_STL_INSTALL_PATH)

        # Remove User config folder if the user requested it
        if os.path.exists(STEAM_STL_CONFIG_PATH) and remove_config:
            print('Removing SteamTInkerLaunch configuration folder...')
            shutil.rmtree(STEAM_STL_CONFIG_PATH)

        # Remove the STL path modification that ProtonUp-Qt may have added during installation from Shell paths
        #
        # Works by getting all the lines in all the hardcoded Shell files that we write out to during installation and
        # and filtering out any line(s) that reference ProtonUp-Qt, then it writes that updated file content back out to the Shell file
        present_shell_files = [
            os.path.join(os.path.expanduser('~'), f) for f in os.listdir(os.path.expanduser('~')) if os.path.isfile(os.path.join(os.path.expanduser('~'), f)) and f in STEAM_STL_SHELL_FILES
        ]
        if os.path.exists(STEAM_STL_FISH_VARIABLES) or shutil.which('fish'):
            present_shell_files.append(STEAM_STL_FISH_VARIABLES)

        print('Removing SteamTinkerLaunch from path...')

        for shell_file in present_shell_files:
            with open(shell_file, 'r+') as mfile:  
                # Get all Shell file lines that are not the ProtonUp-Qt added STL path lines              
                mfile_lines = list(filter(lambda l: 'protonup-qt' not in l.lower() and STEAM_STL_INSTALL_PATH.lower() not in l.lower(), list(mfile.readlines())))
                if len(mfile_lines) == 0:
                    continue
                mfile_lines = mfile_lines[:-1] if len(mfile_lines[-1].strip()) == 0 else mfile_lines

                # Preserve any existing Fish user paths
                if 'fish' in mfile.name:
                    mfile.seek(0)
                    curr_fish_user_paths = list(filter(lambda path: STEAM_STL_INSTALL_PATH not in path, list(get_fish_user_paths(mfile))))
                    updated_fish_user_paths = '\\x1e'.join(curr_fish_user_paths)
                    mfile_lines.append(f'SETUVAR fish_user_paths:{updated_fish_user_paths}')

                # Write out changes while preserving Shell file newlines
                mfile.seek(0)
                prev_line = ''
                for line in mfile_lines:
                    if len(line.strip()) != 0 or len(prev_line.strip()) != 0:
                        mfile.write(line)
                    prev_line = line
                mfile.truncate()

        print('Successfully uninstalled SteamTinkerLaunch!')
        return True
    except IOError as e:
        print('Something went wrong trying to uninstall SteamTinkerLaunch. Aborting...', e)
        return False
